/* (C) 2009-2010,2012 Yana */

#include "device.h"
#include "start_setup.h"
#include "y_base.h"
#include "y_device.h"
#include "y_mips.h"
#include "y_yk0a.h"
#include "yk0a.h"
#include "yk0a_setup.h"
#include <stddef.h>

y_uint8_t task0_stack[TASK0_STACK_SIZE] __attribute__((aligned(Y_STACK_ALIGNMENT)));

y_uint32_t disk_read(y_uint32_t id, y_uint32_t offset, void *address, y_uint32_t size) {
	y_uint32_t i;
	for (i = 0; i < size; ++i) {
		y_uint32_t a;
		if (i % DEVICE_DISK_BUFFER_LEN == 0) {
			if (size - i >= DEVICE_DISK_BUFFER_LEN) {
				a = 0;
			} else {
				if (DEVICE_DISK_BUFFER_LEN - (size - i) > i + offset) {
					a = i + offset;
				} else {
					a = DEVICE_DISK_BUFFER_LEN - (size - i);
				}
			}
			*(volatile y_uint32_t *)DEVICE_DISK_OFFSET = i + offset - a;
			*(volatile y_uint32_t *)DEVICE_DISK_OFFSET_HIGH32 = 0;
			*(volatile y_uint32_t *)DEVICE_DISK_ID = id;
			*(volatile y_uint32_t *)DEVICE_DISK_START_OPERATION = DEVICE_DISK_OPERATION_READ;
			if (*(volatile y_uint32_t *)DEVICE_DISK_STATUS == 0) {
				return i;
			}
		}
		((y_uint8_t *)address)[i] = ((volatile y_uint8_t *)DEVICE_DISK_BUFFER)[i % DEVICE_DISK_BUFFER_LEN + a];
	}
	return size;
}

#include <minix/sys_config.h>

#include <minix/com.h>
#include <minix/const.h>

#include <compat/a.out.h>

#define NAME_SIZE 64

#if !defined(MINIX_KERNEL) || MINIX_KERNEL % CLICK_SIZE != 0
#error MINIX_KERNEL
#endif

extern y_uint8_t _gp[];

struct exec aout[LAST_SPECIAL_PROC_NR + 2];

void task0(void) {
	y_uint32_t a;
	y_uint32_t i;
	y_uint32_t o;
	y_device_cons_initialize();
	a = MINIX_KERNEL;
	o = 0;
	for (i = 0; i < LAST_SPECIAL_PROC_NR + 2; ++i) {
		o += NAME_SIZE;
		if (disk_read(0, o, &aout[i], sizeof(struct exec)) != sizeof(struct exec)) {
			y_device_cons_put_string("{");
			y_device_cons_put_hexadecimal(o);
			y_device_cons_put_string(" ");
			y_device_cons_put_hexadecimal(sizeof(struct exec));
			y_device_cons_put_string("}\n");
			y_device_cons_halt();
		}
		o += sizeof(struct exec);
		aout[i].a_syms = a;
		if (disk_read(0, o, (void *)a, aout[i].a_text) != aout[i].a_text) {
			y_device_cons_put_string("{");
			y_device_cons_put_hexadecimal(o);
			y_device_cons_put_string(" ");
			y_device_cons_put_hexadecimal(aout[i].a_text);
			y_device_cons_put_string("}\n");
			y_device_cons_halt();
		}
		a += aout[i].a_text;
		o += aout[i].a_text;
		if (disk_read(0, o, (void *)a, aout[i].a_data) != aout[i].a_data) {
			y_device_cons_put_string("{");
			y_device_cons_put_hexadecimal(o);
			y_device_cons_put_string(" ");
			y_device_cons_put_hexadecimal(aout[i].a_data);
			y_device_cons_put_string("}\n");
			y_device_cons_halt();
		}
		a = (a + aout[i].a_total + CLICK_SIZE - 1) / CLICK_SIZE * CLICK_SIZE;
		o += aout[i].a_data;
	}
	yk0a_set_priority(1, 1, NULL);
	for (i = 2; i < YK0A_SETUP_TASK_SIZE; ++i) {
		yk0a_connect(i, 1);
		yk0a_set_priority(i, 2, NULL);
	}
	yk0a_set_context(1, YK0A_CONTEXT_GPR4, (yk0a_context_value_t)aout, NULL);
	y_yk0a_start(1, (void (*)(void))aout[0].a_entry, NULL, _gp, y_mips_read_status());
	while (1) {
		yk0a_flag_t f;
		yk0a_task_t t;
		yk0a_listen(YK0A_SETUP_EXCEPTION_PORT, ~YK0A_FLAG_KERNEL_RECEIVER, &f);
		if ((f & YK0A_FLAG_KERNEL_SENDER) != 0 && yk0a_accept(YK0A_SETUP_EXCEPTION_PORT, &t) == 0) {
			if (t != 1) {
				yk0a_forward(1, t);
			} else {
				yk0a_context_value_t cause;
				yk0a_get_context(t, YK0A_CONTEXT_CAUSE, &cause);
				y_device_cons_put_string("{");
				y_device_cons_put_hexadecimal(f);
				y_device_cons_put_string(" ");
				y_device_cons_put_hexadecimal(t);
				y_device_cons_put_string(" ");
				y_device_cons_put_hexadecimal(cause);
				y_device_cons_put_string("}\n");
				y_device_cons_halt();
			}
		}
		if ((f & ~YK0A_FLAG_KERNEL_MASK) != 0) {
			yk0a_bit_flag(1, 0, f & ~YK0A_FLAG_KERNEL_MASK, NULL);
		}
	}
}
